import pytest

import numpy as np

from corgidrp.darks import build_synthesized_dark
from corgidrp.mocks import create_noise_maps, create_dark_calib_files
from corgidrp.detector import unpack_geom, detector_areas, slice_section, embed

im_rows, im_cols, _ = unpack_geom('SCI', 'image', detector_areas)
rows = detector_areas['SCI']['frame_rows']
cols = detector_areas['SCI']['frame_cols']

Fd = np.ones((rows, cols))
Dd = 3/3600*np.ones((rows, cols))
Cd = 0.02*np.ones((rows, cols))

Ferr = np.zeros((rows, cols))
Derr = np.zeros((rows, cols))
Cerr = np.zeros((rows, cols))
Fdq = Ferr.copy().astype(int)
Ddq = Derr.copy().astype(int)
Cdq = Cerr.copy().astype(int)
noise_maps = create_noise_maps(Fd, Ferr, Fdq, Cd,
                                            Cerr, Cdq, Dd, Derr, Ddq)
F = noise_maps.FPN_map
C = noise_maps.CIC_map
D = noise_maps.DC_map
# just need some a dataset for input
np.random.seed(456)
dataset = create_dark_calib_files()
# values used in create_dark_calib_files()
g = 1
t = 60
for fr in dataset.frames:
    fr.ext_hdr['KGAINPAR'] = 7

# convenience function for these tests
def reset_g_t(d, g, t):
    '''
    For a dataset d, reset the EM gain to g and the exposure time to t for
    all frames in d.

    Args:
        d (corgidrp.data.Dataset):  input dataset
        g (float): desired EM gain
        t (float): desired exposure time

    Returns:
        dset (corgidrp.data.Dataset):  dataset with the EM gain and exposure time reset
    '''
    dset = d.copy()
    for fr in dset.frames:
        fr.ext_hdr['EMGAIN_C'] = g
        fr.ext_hdr['EXPTIME'] = t

    return dset

def test_success():
    """Good inputs complete as expected"""
    build_synthesized_dark(dataset, noise_maps)
    build_synthesized_dark(dataset, noise_maps, detector_areas, full_frame=True)
    pass
    
def test_output_size():
    """output is correct size"""
    M = build_synthesized_dark(dataset, noise_maps)
    assert(M.data.shape == (im_rows, im_cols))

def test_exact_case():
    """Exact case produces expected result.  Use full frame this time."""
    tol = 1e-13

    Fd = 5*np.ones((rows, cols))
    Ferr = 0.1*Fd
    Fdq = np.zeros((rows, cols)).astype(int)
    Fdq[3,3] = 1
    Dd = 1/7*np.ones((rows, cols))
    Derr = 0.1*Dd
    Ddq = np.zeros((rows, cols)).astype(int)
    Ddq[3,3] = 1 # Fdq has this one, too
    Ddq[2,2] = 1
    Cd = 1*np.ones((rows, cols))
    Cerr = 0.1*Cd
    Cdq = np.zeros((rows, cols)).astype(int)
    Cdq[1,2] = 1
    g = 5
    t = 7
    target = 3*np.ones((rows, cols))
    exp_err = np.sqrt(Derr**2*g**2*t**2 + Cerr**2*g**2 + Ferr**2)/g
    exp_dq = np.zeros((rows, cols))
    exp_dq[3,3] = 1
    exp_dq[2,2] = 1
    exp_dq[1,2] = 1
    n_maps = create_noise_maps(Fd, Ferr, Fdq, Cd, Cerr, Cdq, Dd, Derr, Ddq)
    dset = reset_g_t(dataset, g, t)

    M = build_synthesized_dark(dset, n_maps, full_frame=True)
    assert(np.max(np.abs(M.data - target)) < tol)
    assert(np.max(np.abs(M.err - exp_err)) < tol)
    assert(np.max(np.abs(M.dq - exp_dq)) < tol)
    assert(M.err_hdr['BUNIT'] == 'Detected Electrons')
    assert(M.ext_hdr['EXPTIME'] == t)
    assert(M.ext_hdr['EMGAIN_C'] == g)
    assert(M.ext_hdr['DATATYPE'] == 'Dark')
    assert(M.data.shape == (rows, cols))
    assert(M.ext_hdr['NAXIS1'] == cols) # NAXIS1 should be cols
    assert(M.ext_hdr['NAXIS2'] == rows)
    assert(M.ext_hdr['DRPNFILE'] ==n_maps.ext_hdr['DRPNFILE']) 
    assert(M.filename == 'Mock0_DRK_CAL.fits')
    assert('EM gain = '+str(g) in str(M.ext_hdr['HISTORY']))
    assert('exptime = '+str(t) in str(M.ext_hdr['HISTORY']))

    M_copy = M.copy()
    assert(np.max(np.abs(M_copy.data - target)) < tol)
    assert(np.max(np.abs(M_copy.err - exp_err)) < tol)
    assert(np.max(np.abs(M_copy.dq - exp_dq)) < tol)
    assert(M_copy.err_hdr['BUNIT'] == 'Detected Electrons')
    assert(M_copy.ext_hdr['EXPTIME'] == t)
    assert(M_copy.ext_hdr['EMGAIN_C'] == g)
    assert(M_copy.ext_hdr['DATATYPE'] == 'Dark')
    assert(M_copy.data.shape == (rows, cols))
    assert(M_copy.ext_hdr['NAXIS1'] == cols) # NAXIS1 should be cols
    assert(M_copy.ext_hdr['NAXIS2'] == rows)
    assert(M_copy.ext_hdr['DRPNFILE'] ==n_maps.ext_hdr['DRPNFILE']) 
    assert(M_copy.filename == 'Mock0_DRK_CAL.fits')
    assert('commanded EM gain = '+str(g) in str(M_copy.ext_hdr['HISTORY']))
    assert('exptime = '+str(t) in str(M_copy.ext_hdr['HISTORY']))
    pass
    
    # test ability to embed an image-area noisemap into full frame, as well as get an image-area master dark
    embedded_maps = []
    for map in [Fd, Cd, Dd]:
        map_slice = slice_section(map, "SCI", 'image', detector_areas)
        # now embed them back into full frames with zeros
        embedded_map = embed(map_slice, 'SCI', 'image', 0)
        embedded_maps.append(embedded_map)
    embedded_maps = np.stack(embedded_maps)
    noise_maps = create_noise_maps(embedded_maps[0], Ferr, Fdq, embedded_maps[1], Cerr, Cdq, 
                        embedded_maps[2], Derr, Ddq)
    dset = reset_g_t(dataset, g, t)
    M_im = build_synthesized_dark(dset, noise_maps, full_frame=False)
    assert(np.max(np.abs(M_im.data - slice_section(target, "SCI", 'image', detector_areas))) < tol)
    assert(np.max(np.abs(M_im.err - slice_section(exp_err, "SCI", 'image', detector_areas))) < tol)
    assert(np.max(np.abs(M_im.dq - slice_section(exp_dq, "SCI", 'image', detector_areas))) < tol)

def test_gain_goes_as_1overg():
    """change in dark goes as 1/g"""
    tol = 1e-13
    noise_maps0 = create_noise_maps(0*Fd, Ferr, Fdq, Cd,
                                        Cerr, Cdq, Dd, Derr, Ddq)
    dset = reset_g_t(dataset, g, t)

    F0 = build_synthesized_dark(reset_g_t(dataset, 1, t), noise_maps0)
    M1 = build_synthesized_dark(reset_g_t(dataset, 1, t), noise_maps)
    M2 = build_synthesized_dark(reset_g_t(dataset, 2, t), noise_maps)
    M4 = build_synthesized_dark(reset_g_t(dataset, 4, t), noise_maps)
    dg1 = M1.data-F0.data
    dg2 = M2.data-F0.data
    dg4 = M4.data-F0.data

    assert(np.max(np.abs(dg2 - dg1/2)) < tol)
    assert(np.max(np.abs(dg4 - dg2/2)) < tol)
    pass

def test_exptime_goes_as_t():
    """change in dark goes as t"""
    tol = 1e-13

    M0 = build_synthesized_dark(reset_g_t(dataset, g, 0), noise_maps)
    M2 = build_synthesized_dark(reset_g_t(dataset, g, 2), noise_maps)
    M4 = build_synthesized_dark(reset_g_t(dataset, g, 4), noise_maps)
    dg2 = M2.data-M0.data
    dg4 = M4.data-M0.data

    assert(np.max(np.abs(dg4 - dg2*2)) < tol)
    pass

def test_c_doesnt_change_with_g_or_t():
    """F = 0 and D = 0 implies C is constant"""
    tol = 1e-13
    noise_maps0 = create_noise_maps(0*Fd, Ferr, Fdq, Cd,
                                        Cerr, Cdq, 0*Dd, Derr, Ddq)
    
    M = build_synthesized_dark(reset_g_t(dataset, 1, t), noise_maps0)
    G2 = build_synthesized_dark(reset_g_t(dataset, 2, t), noise_maps0)
    G4 = build_synthesized_dark(reset_g_t(dataset, 4, t), noise_maps0)
    T2 = build_synthesized_dark(reset_g_t(dataset, g, 2), noise_maps0)
    T4 = build_synthesized_dark(reset_g_t(dataset, g, 4), noise_maps0)
    dg2 = G2.data-M.data
    dg4 = G4.data-M.data
    dt2 = T2.data-M.data
    dt4 = T4.data-M.data

    assert(np.max(np.abs(dg2)) < tol)
    assert(np.max(np.abs(dg4)) < tol)
    assert(np.max(np.abs(dt2)) < tol)
    assert(np.max(np.abs(dt4)) < tol)
    pass

def test_bias_subtracted():
    """check there is no bias when all three noise terms are 0"""
    noise_maps0 = create_noise_maps(0*Fd, Ferr, Fdq, 0*Cd,
                                        Cerr, Cdq, 0*Dd, Derr, Ddq)

    M = build_synthesized_dark(reset_g_t(dataset, 1, t), noise_maps0)
    G2 = build_synthesized_dark(reset_g_t(dataset, 2, t), noise_maps0)
    G4 = build_synthesized_dark(reset_g_t(dataset, 4, t), noise_maps0)
    T2 = build_synthesized_dark(reset_g_t(dataset, g, 2), noise_maps0)
    T4 = build_synthesized_dark(reset_g_t(dataset, g, 4), noise_maps0)

    assert((M.data == 0).all())
    assert((G2.data == 0).all())
    assert((G4.data == 0).all())
    assert((T2.data == 0).all())
    assert((T4.data == 0).all())

    pass

def test_invalid_D_range():
    """Invalid inputs caught as expected"""
    for perr in [-1*np.ones_like(Dd)]:
        nm = create_noise_maps(Fd, Ferr, Fdq, Cd,
                                        Cerr, Cdq, perr, Derr, Ddq)
        with pytest.raises(TypeError):
            build_synthesized_dark(dataset, nm)
        pass
    pass

def test_invalid_C_range():
    """Invalid inputs caught as expected"""
    for perr in [-1*np.ones_like(Cd)]:
        nm = create_noise_maps(Fd, Ferr, Fdq, perr,
                                        Cerr, Cdq, Dd, Derr, Ddq)
        with pytest.raises(TypeError):
            build_synthesized_dark(dataset, nm)
        pass
    pass

def test_g_range_correct():
    """gain is valid >= 1 only"""

    for perr in [-10, -1, 0, 0.999]:
        with pytest.raises(TypeError):
            build_synthesized_dark(reset_g_t(dataset, perr, t), noise_maps)
        pass

    for perr in [1, 1.5, 10]:
        build_synthesized_dark(reset_g_t(dataset, perr, t), noise_maps)
        pass
    pass

if __name__ == '__main__':
    test_success()
    test_output_size()
    test_exact_case()
    test_gain_goes_as_1overg()
    test_exptime_goes_as_t()
    test_c_doesnt_change_with_g_or_t()
    test_bias_subtracted()
    test_invalid_D_range()
    test_invalid_C_range()
    test_g_range_correct()
    pass


import os
import pytest
import numpy as np
import astropy.time as time
from astropy.io import fits
from scipy.signal import decimate
from astropy.modeling import models
from astropy.modeling.models import Gaussian2D

import corgidrp
from corgidrp.mocks import (create_default_L3_headers, create_ct_psfs,
    create_ct_interp, create_ct_cal)
from corgidrp.data import Image, Dataset, FpamFsamCal, CoreThroughputCalibration
from corgidrp import corethroughput

# If a run has crashed before being able to remove the test CT cal file,
# remove it before importing caldb, which scans for new entries in the cal folder
ct_cal_test_file = 'CGI_0200001001001001001_20250415T0305102_CTP_CAL.fits'
if os.path.exists(os.path.join(corgidrp.default_cal_dir, ct_cal_test_file)):
    os.remove(os.path.join(corgidrp.default_cal_dir, ct_cal_test_file))
from corgidrp import caldb

here = os.path.abspath(os.path.dirname(__file__))

def setup_module():
    """
    Create datasets needed for the UTs
    """
    global FPAM_H_CT, FPAM_V_CT, FSAM_H_CT, FSAM_V_CT
    # Choose some H/V values for FPAM/FSAM  during corethroughput observations
    FPAM_H_CT, FPAM_V_CT, FSAM_H_CT, FSAM_V_CT = 6854, 22524, 29471, 12120
    global n_radii, n_azimuths, max_angle
    global cfam_name
    cfam_name = '1F'
    # CT and coronagraphic datasets
    global dataset_ct, dataset_ct_syn, dataset_ct_interp, dataset_psf_eq_rad
    global dataset_cor
    # Arbitrary set of PSF locations to be tested in EXCAM pixels referred to (0,0)
    global psf_loc_in, psf_loc_syn
    global ct_in, ct_syn
   
    # Default headers
    prhd, exthd = create_default_L3_headers()
    # DRP
    exthd['DRPCTIME'] = time.Time.now().isot
    exthd['DRPVERSN'] = corgidrp.__version__
    # cfam filter
    exthd['CFAMNAME'] = cfam_name

    # FPAM/FSAM during CT observing sequence
    exthd['FPAM_H'] = FPAM_H_CT
    exthd['FPAM_V'] = FPAM_V_CT
    exthd['FSAM_H'] = FSAM_H_CT
    exthd['FSAM_V'] = FSAM_V_CT
    data_ct = []
    # Add pupil image(s) of the unocculted source's observation to test that
    # the corethroughput calibration function can handle more than one pupil image
    pupil_image_1 = np.zeros([1024, 1024])
    # Set it to some known value for some known pixels
    pupil_image_1[510:530, 510:530]=1
    pupil_image_2 = np.zeros([1024, 1024])
    # Set it to some known value for some known pixels
    pupil_image_2[510:530, 510:530]=3
    # Add specific values for pupil images:
    # DPAM=PUPIL, LSAM=OPEN, FSAM=OPEN and FPAM=OPEN_12
    exthd_pupil = exthd.copy()
    exthd_pupil['DPAMNAME'] = 'PUPIL'
    exthd_pupil['LSAMNAME'] = 'OPEN'
    exthd_pupil['FSAMNAME'] = 'OPEN'
    exthd_pupil['FPAMNAME'] = 'OPEN_12'
    # Mock error
    err = np.ones([1024,1024])
    # Add pupil images
    dataset_pupil = [Image(pupil_image_1,pri_hdr = prhd, ext_hdr = exthd_pupil, err = err),
        Image(pupil_image_2,pri_hdr = prhd, ext_hdr = exthd_pupil, err = err)]
    data_ct += dataset_pupil
    # Total counts from the pupil images
    unocc_psf_norm = (pupil_image_1.sum()+pupil_image_2.sum())/2
    # Adjust for pupil vs. direct imaging lens transmission
    # This function was written and tested during TVAC tests. There's a test value
    # in test_psf_pix_and_ct() later on
    di_over_pil = corethroughput.di_over_pil_transmission(cfam_name=exthd['CFAMNAME'])
    unocc_psf_norm *= di_over_pil
    # Generate 100 psfs with fwhm=50 mas in band 1 (mock.py)
    data_psf, psf_loc_in, half_psf = create_ct_psfs(50, cfam_name='1F',
        n_psfs=100)
    # Input CT
    ct_in = half_psf/unocc_psf_norm
    # Add PSF images
    data_ct += data_psf
    dataset_ct = Dataset(data_ct)

    # Synthetic PSF for a functional test
    data_ct = []
    psf_test = np.zeros([1024, 1024])
    # Maximum value at one pixel
    psf_loc_syn = [500, 400]
    # Set of known values at selected locations
    psf_test[psf_loc_syn[1]-3:psf_loc_syn[1]+4,
        psf_loc_syn[0]-3:psf_loc_syn[0]+4] = 1
    psf_test[psf_loc_syn[1]-2:psf_loc_syn[1]+3,
        psf_loc_syn[0]-2:psf_loc_syn[0]+3] = 2
    psf_test[psf_loc_syn[1]-1:psf_loc_syn[1]+2,
        psf_loc_syn[0]-1:psf_loc_syn[0]+2] = 3
    psf_test[psf_loc_syn[1],
        psf_loc_syn[0]] = 4
    data_ct += [Image(psf_test,pri_hdr=prhd, ext_hdr=exthd, err=err)]
    # Synthetic pupil images
    data_ct += dataset_pupil
    # Known CT
    ct_syn = (4+3*8+2*16)/(0.5*(1+3)*400)/di_over_pil
    # Dataset
    dataset_ct_syn = Dataset(data_ct) 

    # Coronagraphic dataset (only headers will be used)
    # FPAM/FSAM
    # Choose some H/V values for FPAM/FSAM  during coronagraphic observations
    # These values are *different* than the ones in the dataset_ct defined before
    exthd['FPAM_H'] = FPAM_H_CT - 107
    exthd['FPAM_V'] = FPAM_V_CT + 37
    exthd['FSAM_H'] = FSAM_H_CT + 97
    exthd['FSAM_V'] = FSAM_V_CT - 135
    # FPM center
    exthd['STARLOCX'] = 509
    exthd['STARLOCY'] = 513
    data_cor = [Image(np.zeros([1024, 1024]), pri_hdr=prhd, ext_hdr=exthd, err=err)]
    dataset_cor = Dataset(data_cor)

    # Dataset with some CT profile defined in create_ct_interp
    # Pupil image
    pupil_image = np.zeros([1024, 1024])
    # Set it to some known value for a selected range of pixels
    pupil_image[510:530, 510:530]=1
    prhd, exthd_pupil = create_default_L3_headers()
    # DRP
    exthd_pupil['DRPCTIME'] = time.Time.now().isot
    exthd_pupil['DRPVERSN'] = corgidrp.__version__
    # cfam filter
    exthd_pupil['CFAMNAME'] = cfam_name
    # Add specific values for pupil images:
    # DPAM=PUPIL, LSAM=OPEN, FSAM=OPEN and FPAM=OPEN_12
    exthd_pupil['DPAMNAME'] = 'PUPIL'
    exthd_pupil['LSAMNAME'] = 'OPEN'
    exthd_pupil['FSAMNAME'] = 'OPEN'
    exthd_pupil['FPAMNAME'] = 'OPEN_12'
    # Choose some H/V values for FPAM/FSAM  during corethroughput observations
    exthd_pupil['FPAM_H'] = FPAM_H_CT
    exthd_pupil['FPAM_V'] = FPAM_V_CT
    exthd_pupil['FSAM_H'] = FSAM_H_CT
    exthd_pupil['FSAM_V'] = FSAM_V_CT
    # Mock error
    err = np.ones([1024,1024])
    # Collect Images
    data_ct_interp = [Image(pupil_image,pri_hdr = prhd,
        ext_hdr = exthd_pupil, err = err)]
    # Set of off-axis PSFs with a CT profile defined in create_ct_interp
    # First, we need the CT FPM center to create the CT radial profile
    # We can use a miminal dataset to get to know it
    data_ct_interp += [data_psf[0]]
    ct_cal_tmp = corethroughput.generate_ct_cal(Dataset(data_ct_interp))
    # FPAM/FSAM
    fpam_fsam_cal = FpamFsamCal(os.path.join(corgidrp.default_cal_dir,
        'FpamFsamCal_2024-02-10T00:00:00.000.fits'))
    # FPM during the CT observations (different to the coronagraphic one since
    # FPAM/FSAM H/V values are different)
    fpm_ct = ct_cal_tmp.GetCTFPMPosition(dataset_cor, fpam_fsam_cal)[0]
    # Generate the mock data for CT interpolation knowing the CT FPM
    data_ct_interp = [Image(pupil_image,pri_hdr = prhd,
        ext_hdr = exthd_pupil, err = err)]
    # Synthetic psfs with known CT values (mock.py)
    n_radii = 9
    n_azimuths = 5
    max_angle = 2/3*np.pi
    data_ct_interp += create_ct_interp(
        n_radii=n_radii,
        n_azimuths=n_azimuths,
        max_angle=max_angle,
        fpm_x=fpm_ct[0],
        fpm_y=fpm_ct[1],
        norm=pupil_image_1.sum())[0]
    dataset_ct_interp = Dataset(data_ct_interp)

    # Needed for PSF interpolation
    # Dataset with two PSFs at equal radial distance from the CT FPM's center
    data_psf_tmp = [Image(pupil_image,pri_hdr = prhd,
        ext_hdr = exthd_pupil, err = err)]
    # We need to estimate the location of the FPM's center to create the PSFs
    # at given predefined locatioins
    data_psf_tmp += [data_psf[0]]
    ct_cal_tmp2 = corethroughput.generate_ct_cal(Dataset(data_psf_tmp))
    # FPM during the CT observations (different to the coronagraphic one since
    # FPAM/FSAM H/V values are different)
    fpm_ct_2 = ct_cal_tmp2.GetCTFPMPosition(dataset_cor, fpam_fsam_cal)[0]
    # Generate the mock data for CT interpolation knowing the CT FPM
    data_psf_eq_rad = [Image(pupil_image,pri_hdr = prhd,
        ext_hdr = exthd_pupil, err = err)]
    # Band 1 FWHM: Enough approximation (and to a good extent, irrelevant)
    fwhm_mas = 50
    imshape = (219, 219)
    y, x = np.indices(imshape)
    # Following astropy documentation:
    # Generate 2 PSFs with the same radial distance to (0,0) and different A
    fpm_ct_frac_2 = fpm_ct_2 % 1
    # Choose the location of the PSFs
    aa = 100
    bb = 20
    model_params = []
    # Fill out the four quadrants to have more of a variety with PSFs at equal radii
    for idx in range(4):
        aa_w_sgn = (-1)**idx * aa
        bb_w_sgn = (-1)**(idx//2) * bb
        model_params += [
            dict(amplitude=11,
                x_mean=imshape[0]//2 + fpm_ct_frac_2[0]+aa_w_sgn,
                y_mean=imshape[1]//2 + fpm_ct_frac_2[1]+bb_w_sgn,
                x_stddev=fwhm_mas/21.8/2.335,
                y_stddev=fwhm_mas/21.8/2.335),
            dict(amplitude=23,
                x_mean=imshape[0]//2 + fpm_ct_frac_2[0]+bb_w_sgn,
                y_mean=imshape[1]//2 + fpm_ct_frac_2[1]+aa_w_sgn,
                x_stddev=fwhm_mas/21.8/2.335,
                y_stddev=fwhm_mas/21.8/2.335)]
    model_list = [models.Gaussian2D(**kwargs) for kwargs in model_params]
    # Render models to image using full evaluation
    for model in model_list:
        psf = np.zeros(imshape)
        model.bounding_box = None
        model.render(psf)
        image = np.zeros([1024, 1024])
        # Insert PSF 
        image[int(fpm_ct_2[1])-imshape[1]//2:int(fpm_ct_2[1])+imshape[1]//2+1,
            int(fpm_ct_2[0])-imshape[0]//2:int(fpm_ct_2[0])+imshape[0]//2+1] = psf
        # Build up the Dataset
        data_psf_eq_rad += [Image(image,pri_hdr=prhd, ext_hdr=exthd, err=err)]    

    dataset_psf_eq_rad = Dataset(data_psf_eq_rad)

def test_psf_pix_and_ct():
    """
    Test 1090881 - Given a core throughput dataset consisting of M clean frames
    (nominally 1024x1024) taken at different FSM locations, the CTC GSW shall
    estimate the pixel location and core throughput of each PSF.

    NOTE: the list of M clean frames may be a subset of the frames collected during
    core throughput data collection, to allow for the removal of outliers.
    """
    # Test 1:
    # Check that the step function retrieves the expected location and CT of
    # a set of simulated 2D Gaussian PSFs (created in setup_module before:)
    psf_loc_est, ct_est = corethroughput.estimate_psf_pix_and_ct(dataset_ct)
    # Difference between expected and retrieved locations for the max (peak) method
    diff_psf_loc = psf_loc_in - psf_loc_est
    # Set a difference of 0.005 pixels
    assert np.all(np.abs(diff_psf_loc) <= 0.005)
    # core throughput in (0,1]
    assert np.all(ct_est) > 0
    assert np.all(ct_est) <= 1
    # comparison between I/O values (<=1% due to pixelization effects vs.
    # expected analytical value)
    assert np.all(np.abs(ct_est-ct_in) <= 0.01)

    # Test 2:
    # Functional test with some mock data with known PSF location and CT
    # Synthetic PSF from setup_module
    psf_loc_est, ct_est = corethroughput.estimate_psf_pix_and_ct(dataset_ct_syn)
    # In this test, there must be an exact agreement
    assert np.all(psf_loc_est[0] == psf_loc_syn)
    assert np.abs(ct_est[0]-ct_syn) < 1e-16

    print('Tests about PSF locations and CT values passed')

def test_fpm_pos():
    """
    Test 1090882 - Given 1) the location of the center of the FPM coronagraphic
    mask in EXCAM pixels during the coronagraphic observing sequence and 2) the
    FPAM and FSAM encoder positions during both the coronagraphic and core
    throughput observing sequences, the CTC GSW shall compute the center of the
    FPM coronagraphic mask during the core throughput observing sequence.

    Python algorithm written during TVAC for using FPAM and FSAM matrices
    
          delta_pam = np.array([[dh], [dv]]) # fill these in
          # read FPAM or FSAM matrices: 
          M = np.array([[ M00, M01], [M10, M11]], dtype=float32)
          delta_pix = M @ delta_pam
    """
    # FPAM/FSAM transformations
    fpam_fsam_cal = FpamFsamCal(os.path.join(corgidrp.default_cal_dir,
        'FpamFsamCal_2024-02-10T00:00:00.000.fits'))
    
    fpam2excam_matrix, fsam2excam_matrix = fpam_fsam_cal.data
    # TVAC files
    fpam2excam_matrix_tvac = fits.getdata(os.path.join(here, 'test_data',
        'fpam_to_excam_modelbased.fits'))
    fsam2excam_matrix_tvac = fits.getdata(os.path.join(here, 'test_data',
        'fsam_to_excam_modelbased.fits'))

    # test 1:
    # Check that DRP calibration files for FPAM and FSAM agree with TVAC files
    # Some irrelevant rounding happens when defining FpamFsamCal() in data.py
    assert np.all(fpam2excam_matrix - fpam2excam_matrix_tvac <= 1e-9)
    assert np.all(fsam2excam_matrix - fsam2excam_matrix_tvac <= 1e-9)
   
    # test 2:
    # Using values within the range should return a meaningful value. Tested 10 times
    FPM_center_pos_pix = np.array([dataset_cor[0].ext_hdr['STARLOCX'],
            dataset_cor[0].ext_hdr['STARLOCY']])
    FPAM_center_pos_um = np.array([dataset_cor[0].ext_hdr['FPAM_H'],
            dataset_cor[0].ext_hdr['FPAM_V']])
    FSAM_center_pos_um =  np.array([dataset_cor[0].ext_hdr['FSAM_H'],
            dataset_cor[0].ext_hdr['FSAM_V']])
    rng = np.random.default_rng(0)
    for _ in range(10):
        # Random shift for Delta H/V in um
        delta_fpam_um = np.array([rng.uniform(1,10), rng.uniform(1,10)])
        delta_fsam_um = np.array([rng.uniform(1,10), rng.uniform(1,10)])
        # Update dataset_ct headers
        dataset_ct[0].ext_hdr['FPAM_H'] = dataset_cor[0].ext_hdr['FPAM_H'] + delta_fpam_um[0]
        dataset_ct[0].ext_hdr['FPAM_V'] = dataset_cor[0].ext_hdr['FPAM_V'] + delta_fpam_um[1]
        dataset_ct[0].ext_hdr['FSAM_H'] = dataset_cor[0].ext_hdr['FSAM_H'] + delta_fsam_um[0]
        dataset_ct[0].ext_hdr['FSAM_V'] = dataset_cor[0].ext_hdr['FSAM_V'] + delta_fsam_um[1]
        # Create CT cal file
        ct_cal_tmp = corethroughput.generate_ct_cal(dataset_ct)
        # Get CT FPM center
        fpam_ct_pix_out, fsam_ct_pix_out = \
            ct_cal_tmp.GetCTFPMPosition(
                dataset_cor,
                fpam_fsam_cal)
        # Expected shifts in EXCAM pixels
        delta_fpam_pix = fpam2excam_matrix @ delta_fpam_um
        delta_fsam_pix = fsam2excam_matrix @ delta_fsam_um
        # Compare output and input: Some of the random tests have differences
        # of ~1e-13, while others are exactly equal
        fpam_ct_pix_in = FPM_center_pos_pix + delta_fpam_pix
        assert np.all(fpam_ct_pix_out - fpam_ct_pix_in <= 1e-12)
        fsam_ct_pix_in = FPM_center_pos_pix + delta_fsam_pix
        assert np.all(fsam_ct_pix_out - fsam_ct_pix_in <= 1e-12)

    print('Tests about FPAM/FSAM to EXCAM passed')

def test_cal_file():
    """ Test creation of core throughput calibration file. """

    # Write core throughput calibration file
    ct_cal_file_in = corethroughput.generate_ct_cal(dataset_ct)
    ct_cal_file_in.save(filedir=corgidrp.default_cal_dir)

    # Check that the filename is what we expect
    ct_cal_filename = dataset_ct[-1].filename.replace("_L2b", "_CTP_CAL")
    ct_cal_filepath = os.path.join(corgidrp.default_cal_dir,ct_cal_filename)
    if os.path.exists(ct_cal_filepath) is False:
        raise IOError(f'Core throughput calibration file {ct_cal_filepath} does not exist.')
    # Load the calibration file to check it has the same data contents
    ct_cal_file_load = CoreThroughputCalibration(ct_cal_filepath)
    assert np.all(ct_cal_file_load.data == ct_cal_file_in.data)
    assert np.all(ct_cal_file_load.err == ct_cal_file_in.err)
    assert np.all(ct_cal_file_load.dq == ct_cal_file_in.dq)
    assert np.all(ct_cal_file_load.ct_excam == ct_cal_file_in.ct_excam)
    assert np.all(ct_cal_file_load.ct_fpam == ct_cal_file_in.ct_fpam)
    assert np.all(ct_cal_file_load.ct_fsam == ct_cal_file_in.ct_fsam)
    
    # This test checks that the CT cal file has the right information by making
    # sure that I=O (Note: the comparison b/w analytical predictions
    # vs. centroid/pixelized data is part of the tests on test_psf_pix_and_ct before)

    # Input values for PSF centers and corresponding CT values
    psf_loc_input, ct_input = \
        corethroughput.estimate_psf_pix_and_ct(dataset_ct)

    # Test: open calibration file
    try:
        ct_cal_file = corgidrp.data.CoreThroughputCalibration(ct_cal_file_in.filepath)
    except:
        raise IOError('CT cal file was not saved')

    # Test 1: PSF cube
    # Recover off-axis PSF cube from CT Dataset
    psf_cube_in = []
    for frame in dataset_ct:
        try:
        # Pupil images of the unocculted source satisfy:
        # DPAM=PUPIL, LSAM=OPEN, FSAM=OPEN and FPAM=OPEN_12
            exthd = frame.ext_hdr
            if (exthd['DPAMNAME']=='PUPIL' and exthd['LSAMNAME']=='OPEN' and
                exthd['FSAMNAME']=='OPEN' and exthd['FPAMNAME']=='OPEN_12'):
                continue
        except:
           pass
        psf_cube_in += [frame.data]
    psf_cube_in = np.array(psf_cube_in)

    # Compare the PSF cube from the calibration file, which may have a smaller
    # extension, with the input ones
    cal_file_side_0 = ct_cal_file.data.shape[1]
    cal_file_side_1 = ct_cal_file.data.shape[2]
    for i_psf, psf in enumerate(psf_cube_in):
        loc_00 = np.argwhere(psf == ct_cal_file.data[i_psf][0][0])[0]
        assert np.all(psf[loc_00[0]:loc_00[0]+cal_file_side_0,
            loc_00[1]:loc_00[1]+cal_file_side_1] == ct_cal_file.data[i_psf])

    # Test 2: PSF positions and CT map
    # Check EXTNAME is as expected
    if ct_cal_file.ct_excam_hdr['EXTNAME'] != 'CTEXCAM':
        raise ValueError('The extension name of the CT values on EXCAM is not correct')
    # x location wrt FPM
    assert np.all(psf_loc_input[:,0] == ct_cal_file.ct_excam[0])
    # y location wrt FPM
    assert np.all(psf_loc_input[:,1] == ct_cal_file.ct_excam[1])
    # CT map
    assert np.all(ct_input == ct_cal_file.ct_excam[2])

    # Test 3: FPAM and FSAM positions during the CT observations are present
    # Check EXTNAME is as expected
    if ct_cal_file.ct_fpam_hdr['EXTNAME'] != 'CTFPAM':
        raise ValueError('The extension name of the FPAM values on EXCAM is not correct')
    assert np.all(ct_cal_file.ct_fpam == np.array([dataset_ct[0].ext_hdr['FPAM_H'],
        dataset_ct[0].ext_hdr['FPAM_V']]))
    if ct_cal_file.ct_fsam_hdr['EXTNAME'] != 'CTFSAM':
        raise ValueError('The extension name of the FSAM values on EXCAM is not correct')
    assert np.all(ct_cal_file.ct_fsam == np.array([dataset_ct[0].ext_hdr['FSAM_H'],
        dataset_ct[0].ext_hdr['FSAM_V']]))
    
    # Remove test CT cal file
    if os.path.exists(ct_cal_file.filepath):
        os.remove(ct_cal_file.filepath)

    print('Tests about the CT cal file passed')

def test_ct_interp():
    """ Tests the interpolation within the standard range by popping out data
    points and checking that the interpolation is < 5% error. The core
    throughput changes linearly across the radius.
    """

    # Generate core throughput calibration file
    ct_cal_in = corethroughput.generate_ct_cal(dataset_ct_interp)
    # Get CT FPM center
    # FPAM/FSAM
    fpam_fsam_cal = FpamFsamCal(os.path.join(corgidrp.default_cal_dir,
        'FpamFsamCal_2024-02-10T00:00:00.000.fits'))
    fpam_ct_pix = ct_cal_in.GetCTFPMPosition(dataset_cor, fpam_fsam_cal)[0]
    # Reference grid to test the interpolation: wrt CT FPM because the positions
    # used to interpolate the CT are, by definition, wrt the FPM
    x_grid = ct_cal_in.ct_excam[0,:] - fpam_ct_pix[0]
    y_grid = ct_cal_in.ct_excam[1,:] - fpam_ct_pix[1]
    core_throughput = ct_cal_in.ct_excam[2,:]
    
    # In this test, we will estimate the CT at a location that agrees with one
    # of the locations used to build the CT interpolation set by removing
    # the data point and estimating the CT with the remaining unchanged set

    # Remember that all the datasets will share the same pupil image. Only one
    # off-axis PSF is removed
    pupil_image = np.zeros([1024, 1024])
    # Set it to some known value for a selected range of pixels
    pupil_image[510:530, 510:530]=1
    prhd, exthd_pupil = create_default_L3_headers()
    # DRP
    exthd_pupil['DRPCTIME'] = time.Time.now().isot
    exthd_pupil['DRPVERSN'] = corgidrp.__version__
    # cfam filter
    exthd_pupil['CFAMNAME'] = cfam_name
    # Add specific values for pupil images:
    # DPAM=PUPIL, LSAM=OPEN, FSAM=OPEN and FPAM=OPEN_12
    exthd_pupil['DPAMNAME'] = 'PUPIL'
    exthd_pupil['LSAMNAME'] = 'OPEN'
    exthd_pupil['FSAMNAME'] = 'OPEN'
    exthd_pupil['FPAMNAME'] = 'OPEN_12'
    # Choose some H/V values for FPAM/FSAM  during corethroughput observations
    # Remember the dataset_cor used in InterpolateCT() below has different values
    # to simulate different values of the FPM center during coronagraphic and
    # core throughput observing sequences
    exthd_pupil['FPAM_H'] = FPAM_H_CT
    exthd_pupil['FPAM_V'] = FPAM_V_CT
    exthd_pupil['FSAM_H'] = FSAM_H_CT
    exthd_pupil['FSAM_V'] = FSAM_V_CT
    # Mock error
    err = np.ones([1024,1024])
    # Generate random indices between 0 and the number of radii and azimuths,
    # excluding the edge cases 
    n_random = 50
    # Set seed for reproducibility of test data
    rng = np.random.default_rng(0)
    for idx in range(n_random):
        random_index_radius = rng.choice(np.arange(1, n_radii-1), 1)
        random_index_az = rng.choice(np.arange(1, n_azimuths-1), 1)
     
        #Convert these to flattned indices
        random_indices_flat = random_index_radius + random_index_az*n_radii
        
        # Record the missing value
        missing_x = x_grid[random_indices_flat]
        missing_y = y_grid[random_indices_flat]
        missing_core_throughput = core_throughput[random_indices_flat]
        # Generate CT dataset w/o the latter (needed to call the interpolant
        # without this location)
        # Dataset for CT map interpolation: pupil images plus off-axis PSFs
        data_ct = [Image(pupil_image,pri_hdr = prhd, ext_hdr = exthd_pupil, err = err)]
        data_ct += create_ct_interp(
            n_radii=n_radii,
            n_azimuths=n_azimuths,
            max_angle=max_angle,
            norm=pupil_image.sum(),
            fpm_x=fpam_ct_pix[0],
            fpm_y=fpam_ct_pix[1],
            pop_index=random_indices_flat)[0]
        dataset_ct_tmp = Dataset(data_ct)
        # Generate core throughput calibration file
        ct_cal_tmp = corethroughput.generate_ct_cal(dataset_ct_tmp)
        # Now we can interpolate the missing values
        # Test with linear mapping of radii
        interpolated_value = ct_cal_tmp.InterpolateCT(
            missing_x, missing_y, dataset_cor, fpam_fsam_cal, logr=False)[0]
        # Good to within 2% 
        assert interpolated_value == pytest.approx(missing_core_throughput, abs=0.05), 'Error more than 5% (linear radii mapping)'
        # Test with radii mapped into their logarithmic values before
        # constructing the interpolant 
        interpolated_value_log = ct_cal_tmp.InterpolateCT(
            missing_x, missing_y, dataset_cor, fpam_fsam_cal, logr=True)[0]
        # Good to within 2%
        assert interpolated_value_log == pytest.approx(missing_core_throughput, abs=0.05), 'Error more than 5% (logarithmic radii mapping)'
        
    # Test that if the radius is out of the range then an error is thrown
    # Pick a data point that is out of the range. For instance, set y to zero
    # and x to a value that is greater than the maximum radius
    radii = np.sqrt(x_grid**2 + y_grid**2)
    with pytest.raises(ValueError):
        # Too Big
        ct_cal_tmp.InterpolateCT(radii.max()+1, 0, dataset_cor, fpam_fsam_cal) 
             
    with pytest.raises(ValueError):
        #Too small
        ct_cal_tmp.InterpolateCT(0.9*radii.min(), 0, dataset_cor, fpam_fsam_cal)

    # Test that something with an azimuth out of range returns the same result
    # as within the range
    azimuths = np.arctan2(y_grid, x_grid)
    azimuths -= azimuths.min()
    x_new_out = 0.9*np.max(radii)*np.cos(np.max(azimuths)+0.1)
    y_new_out = 0.9*np.max(radii)*np.sin(np.max(azimuths)+0.1)
    interpolated_value_out = ct_cal_tmp.InterpolateCT(
        x_new_out, y_new_out, dataset_cor, fpam_fsam_cal)[0]

    x_new_in = 0.9*np.max(radii)*np.cos(0.1)
    y_new_in = 0.9*np.max(radii)*np.sin(0.1)
    interpolated_value_in = ct_cal_tmp.InterpolateCT(
        x_new_in, y_new_in, dataset_cor, fpam_fsam_cal)[0]

    assert interpolated_value_out == pytest.approx(interpolated_value_in, abs=0.01), "Error more than 1% error"
    # Make sure it still works with a non-zero starting azimuth: min_angle below
    data_ct = [Image(pupil_image,pri_hdr = prhd, ext_hdr = exthd_pupil,
            err = err)]
    data_ct_interp = create_ct_interp(
        n_radii=n_radii,
        n_azimuths=n_azimuths,
        min_angle=-0.1,
        max_angle=max_angle,
        fpm_x=fpam_ct_pix[0],
        fpm_y=fpam_ct_pix[1],
        norm=pupil_image.sum())[0]
    data_ct += data_ct_interp
    dataset_ct_az = Dataset(data_ct)
    # Generate core throughput calibration file
    ct_cal_az = corethroughput.generate_ct_cal(dataset_ct_az)

    # Out of range of the new shifted azimuths
    x_az_out = 0.9*np.max(radii) * np.cos(max_angle + 0.2)
    y_az_out = 0.9*np.max(radii) * np.sin(max_angle + 0.2)
    interpolated_value_az_out = ct_cal_az.InterpolateCT(
        x_az_out, y_az_out, dataset_cor, fpam_fsam_cal)[0]

    # In range of the new shifted azimuths
    x_az_in = 0.9*np.max(radii) * np.cos(0.2)
    y_az_in = 0.9*np.max(radii) * np.sin(0.2)
    interpolated_value_az_in = ct_cal_az.InterpolateCT(
        x_az_in, y_az_in, dataset_cor, fpam_fsam_cal)[0]
    
    assert interpolated_value_az_out == pytest.approx(interpolated_value_az_in, abs=0.01), "Error more than 1% error"

    print('Tests about CT interpolation passed')

def test_get_1d_ct():
    """Test that corethroughput.get_1d_ct() produces an array of the correct 
    shape and returns the expected PSF for each position."""

    d = 2.36 #m
    lam = 573.8e-9 #m
    pixscale_arcsec = 0.0218
    fwhm_mas = 1.22 * lam / d * 206265 * 1000
    
    # Test where DETPIX0X/Y = (0,0)
    nx,ny = (5,5)
    cenx, ceny = (25.5,30.5)
    ctcal = create_ct_cal(fwhm_mas,
                  cenx=cenx,ceny=ceny,
                  nx=nx,ny=ny)
    
    pri_hdr = fits.Header()
    ext_hdr = fits.Header()
    ext_hdr["STARLOCX"] = 25.
    ext_hdr["STARLOCY"] = 30.
    frame = Image(np.zeros([80,80]),
                  pri_hdr=pri_hdr,
                  ext_hdr=ext_hdr)
    
    seps = [0.,1.,1.41,2.,3.,4.]

    expected_args = [12,7,6,2,0,0]

    ct_1d = corethroughput.get_1d_ct(ctcal,frame,
                                     seps)
    
    assert ct_1d.shape == (2,len(seps))

    for i,arg in enumerate(expected_args):
        assert ct_1d[1,i] == ctcal.ct_excam[2,arg]

    # Test where DETPIX0X/Y is nonzero:

    nx,ny = (5,5)
    cenx, ceny = (45.5,35.5)
    ctcal = create_ct_cal(fwhm_mas,
                  cenx=cenx,ceny=ceny,
                  nx=nx,ny=ny)
    
    pri_hdr = fits.Header()
    ext_hdr = fits.Header()
    ext_hdr["STARLOCX"] = 25.
    ext_hdr["STARLOCY"] = 30.
    ext_hdr["DETPIX0X"] = 20
    ext_hdr["DETPIX0Y"] = 5
    frame = Image(np.zeros([80,80]),
                  pri_hdr=pri_hdr,
                  ext_hdr=ext_hdr)
    
    seps = [0.,1.,1.41,2.,3.,4.]

    expected_args = [12,7,6,2,0,0]

    ct_1d = corethroughput.get_1d_ct(ctcal,frame,
                                     seps)
    
    assert ct_1d.shape == (2,len(seps))

    for i,arg in enumerate(expected_args):
        assert ct_1d[1,i] == ctcal.ct_excam[2,arg]


def test_ct_map():
    """ Tests the creation of a core throughput map. The method InterpolateCT()
      has its own unit test and can be considered as tested in the following. """

    # I run the test for two completely different CT datasets: 2D PSFs
    # randomly distributed (dataset_ct) and the one used for CT interpolation
    # that has a set of PSFs with pre-stablished CT and locations (dataset_ct_interp)

    for dataset in [dataset_ct, dataset_ct_interp]:
        # Generate core throughput calibration file
        ct_cal = corethroughput.generate_ct_cal(dataset)
    
        # FPAM/FSAM
        fpam_fsam_cal = FpamFsamCal(os.path.join(corgidrp.default_cal_dir,
            'FpamFsamCal_2024-02-10T00:00:00.000.fits'))
    
        # Create CT map for the HLC area (default)
        ct_map_def = corethroughput.CreateCTMap(dataset_cor, fpam_fsam_cal, ct_cal)

        # CT values are within [0,1]
        assert ct_map_def[2].min() >= 0
        assert ct_map_def[2].max() <= 1
        # Verify CT values are within the range of the input CT values
        # Allow some minimum tolerance due to float64 numerical precision
        tolerance = 1e-14
        assert ct_map_def[2].min() >= ct_cal.ct_excam[2].min() - tolerance
        assert ct_map_def[2].max() <= ct_cal.ct_excam[2].max() + tolerance

        # Additional test to compate the CT map with the expected model from
        # create_ct_interp (predefined to be radial and linear). The other
        # dataset, dataset_ct is not suitable for this specific test because
        # it's made of random 2D Gaussians
        if dataset == dataset_ct_interp:
            r_def = np.sqrt(ct_map_def[0]**2 + ct_map_def[1]**2)
            ct_def = r_def/r_def.max()
            # create_ct_interp did not include the pupil lens to imaging lens ratio
            ct_def /= corethroughput.di_over_pil_transmission(cfam_name)
  
            # Differences below 1% are good
            assert ct_map_def[2] == pytest.approx(ct_def, abs=0.01), 'Differences are greater than 1%' 
    
        # Test the ability to parse some user-defined locations
        # If the target pixels are the same as the ones in the CT file, the
        # locations *and* CT values must agree with the ones in the CT file
        # Get FPM's center during CT observations
        ct_fpm = ct_cal.GetCTFPMPosition(dataset_cor, fpam_fsam_cal)[0]
        target_pix = [ct_cal.ct_excam[0] - ct_fpm[0],
            ct_cal.ct_excam[1] - ct_fpm[1]]
        ct_map_targ = corethroughput.CreateCTMap(dataset_cor, fpam_fsam_cal, ct_cal,
            target_pix=target_pix)
    
        # All locations must have a valid CT value
        assert target_pix[0] == pytest.approx(ct_map_targ[0], abs=1e-14)
        assert target_pix[1] == pytest.approx(ct_map_targ[1], abs=1e-14)
        # CT values must be the same
        assert ct_cal.ct_excam[2] == pytest.approx(ct_map_targ[2], abs=1e-14)

    # Test it can be saved
    default_filepath = os.path.join(corgidrp.default_cal_dir, 'ct_map.csv')
    ct_map_def = corethroughput.CreateCTMap(dataset_cor, fpam_fsam_cal, ct_cal,
        save = True)
    assert os.path.exists(default_filepath), f"File not found: {default_filepath}"

    # Add open the file and compare content
    ct_map_saved = np.loadtxt(default_filepath, delimiter=',')
    assert np.all(ct_map_saved == ct_map_def)

    print('Tests about CT map passed')

def test_psf_interp():
    """
    Test the ability to recover a PSF at a given (x,y) location on HLC in a
    coronagraphic observation given a CT calibration file and the PAM
    transformation from encoder values to EXCAM pixels.
    """
    # Test 1/ Equal radial distance should retrieve the one with the nearest
    # angular distance
    # Generate core throughput calibration file
    ct_cal_eq = corethroughput.generate_ct_cal(dataset_psf_eq_rad)

    # We need to refer the PSF locations to the FPM's center
    # Get PAM cal file
    fpam_fsam_cal = FpamFsamCal(os.path.join(corgidrp.default_cal_dir,
        'FpamFsamCal_2024-02-10T00:00:00.000.fits'))
    # Get CT FPM's center
    fpam_ct_pix_eq = ct_cal_eq.GetCTFPMPosition(dataset_cor, fpam_fsam_cal)[0]
    # PSF locations with respect to the FPM's center. EXCAM pixels
    x_ct_eq = ct_cal_eq.ct_excam[0] - fpam_ct_pix_eq[0]
    y_ct_eq = ct_cal_eq.ct_excam[1] - fpam_ct_pix_eq[1]
    r_ct_eq = np.sqrt(x_ct_eq**2 + y_ct_eq**2)

    # Test with a location with the same radius and an equally (absolute) angular
    # distance from two PSFs in the input dataset: It should be the average value
    # In any interpolation beyind the nearest, there will be some
    # weighting factor that takes into account the distances
    x_mid = np.sqrt(100**2+20**2)/np.sqrt(2)
    y_mid = x_mid
    psf_interp = ct_cal_eq.GetPSF(x_mid, y_mid, dataset_cor, fpam_fsam_cal)[0]
    assert np.all(psf_interp == 0.5*(ct_cal_eq.data[0]+ct_cal_eq.data[1]))

    # Test with target location with the same radius as two PSFs in the input
    # dataset, but at a different angular location closer to one of them
    # Closer to the 1st one (arbitrary small shift, one of many possible choices)
    x_targ = r_ct_eq[0]*np.cos(np.arctan(y_ct_eq[0]/x_ct_eq[0])-0.05)
    y_targ = r_ct_eq[0]*np.sin(np.arctan(y_ct_eq[0]/x_ct_eq[0])-0.05)
    psf_interp = ct_cal_eq.GetPSF(x_targ, y_targ, dataset_cor, fpam_fsam_cal)[0]
    assert np.all(psf_interp == ct_cal_eq.data[0])
   
    # Closer to the 2nd one (arbitrary small shift, one of many possible choices)
    x_targ = r_ct_eq[1]*np.cos(np.arctan(y_ct_eq[1]/x_ct_eq[1])+0.06)
    y_targ = r_ct_eq[1]*np.sin(np.arctan(y_ct_eq[1]/x_ct_eq[1])+0.06)
    psf_interp = ct_cal_eq.GetPSF(x_targ, y_targ, dataset_cor, fpam_fsam_cal)[0]
    assert np.all(psf_interp == ct_cal_eq.data[1])
    
    # Next tests: Generate core throughput calibration file with more PSFs
    ct_cal = corethroughput.generate_ct_cal(dataset_ct_interp)

    # We need to refer the PSF locations to the FPM's center
    # Get PAM cal file
    fpam_fsam_cal = FpamFsamCal(os.path.join(corgidrp.default_cal_dir,
        'FpamFsamCal_2024-02-10T00:00:00.000.fits'))
    # Get CT FPM's center
    fpam_ct_pix = ct_cal.GetCTFPMPosition(dataset_cor, fpam_fsam_cal)[0]
    # PSF locations with respect to the FPM's center. EXCAM pixels
    x_ct = ct_cal.ct_excam[0,:] - fpam_ct_pix[0]
    y_ct = ct_cal.ct_excam[1,:] - fpam_ct_pix[1]
    # Radial distance
    r_ct = np.sqrt(x_ct**2 + y_ct**2)

    # Number of positions to test (PSF stamps are small 15x15 pixels)
    n_rand = 200
    rng = np.random.default_rng(0)

    # Test 2: Check that if the location is below or above the range of radial distances
    # in the input CT dataset used to generate the CT cal file, it fails
    # Test 2.1/ Below minimum radial distance:
    x_below = rng.uniform(0, r_ct.min()-0.2, n_rand)
    y_below = r_ct.min() - x_below 
    # All must fail
    with pytest.raises(ValueError):
        ct_cal.GetPSF(x_below, y_below, dataset_cor, fpam_fsam_cal)
    # Test 2.2/ Beyond maximum radial distance 
    x_beyond = rng.uniform(r_ct.max()+0.2, 2*r_ct.max(), n_rand)
    y_beyond = x_beyond - r_ct.max()
    # All must fail
    with pytest.raises(ValueError):
        ct_cal.GetPSF(x_beyond, y_beyond, dataset_cor, fpam_fsam_cal)

    # Test 3/ Choose some arbitrary positions and check the returned PSF is the
    # same as the nearest one in the CT cal file:
    
    # Array of random numbers, covering all quadrants with radial distances from
    # the FPM within the range of the CT cal file including some locations that
    # are invalid
    r_test = rng.uniform(r_ct.min()-1, r_ct.max()+1, n_rand)
    az_test = rng.uniform(0, 2*np.pi, n_rand)
    x_test = r_test * np.cos(az_test)
    y_test = r_test * np.sin(az_test)
    # Number of locations outside the range of valid radial distances
    n_invalid = ((r_test<r_ct.min()) + (r_test>r_ct.max())).sum()
    if n_invalid == 0:
        print('Target dataset should have some locations outside the radial range of the input dataset')

    # Interpolated PSFs
    interpolated_PSF, x_out, y_out = ct_cal.GetPSF(x_test, y_test, dataset_cor, fpam_fsam_cal)
    # Test that the number of interpolated PSFs is the same as the number of
    # valid positions
    if n_invalid + len(interpolated_PSF) != n_rand:
        raise Exception('Inconsistent number of interpolated PSFs')

    # Radial distance
    r_ct = np.sqrt(x_ct**2 + y_ct**2)
    for i_psf in range(len(x_out)):
        # Radial distance difference with the input dataset
        r_out = np.sqrt(x_out[i_psf]**2 + y_out[i_psf]**2)
        # Test agreement between interpolated PSFs and nearest one in the input set
        # Remember agreement to bin radial distances to 1/10th of a pixel in the
        # nearest-polar method
        diff_r_abs = np.round(10*np.abs(r_out - r_ct)/10)
        idx_near = np.argwhere(diff_r_abs == diff_r_abs.min())
        # If there's more than one case, check the interpolated PSF is the one
        # that has the shortest angular distance to the ones in the input dataset
        # with the same radial distance, or if there are two such locations
        # (half angle), the output should be the average of both PSFs (agreement)
        if len(idx_near) > 1:
            # Difference in angle b/w target and grid
            # We want to distinguish PSFs at different quadrants
            az_grid = np.arctan2(y_ct[idx_near], x_ct[idx_near])
            az_cor = np.arctan2(y_out[i_psf], x_out[i_psf])
            # Flatten into a 1-D array
            diff_az_abs = np.abs(az_cor - az_grid).transpose()[0]
            # Azimuth binning consistent with the binning of the radial distance
            bin_az_fac = 1/10/r_out
            diff_az_abs = bin_az_fac * np.round(diff_az_abs/bin_az_fac)
            # Closest angular location to the target location within equal radius
            idx_near_az = np.argwhere(diff_az_abs == diff_az_abs.min())
            # If there are two locations (half angle), choose the average (agreement)
            if len(idx_near_az) == 2:
                assert np.all(interpolated_PSF[i_psf] ==
                    ct_cal.data[idx_near[idx_near_az]].mean(axis=0))
            # Otherwise, this is the PSF
            elif len(idx_near_az) == 1:
                assert np.all(interpolated_PSF[i_psf] ==
                    ct_cal.data[idx_near[idx_near_az[0]]])
            else:
                raise ValueError(f'There are {len(idx_near_az):d} PSFs ',
                            'equally near the target PSF. This should not happen.')
        # Otherwise this is the interpolated PSF
        else:
            assert np.all(interpolated_PSF[i_psf] == ct_cal.data[idx_near[0]])

    print('Tests about PSF interpolation passed')

def teardown_module():
    """
    Deletes variables
    """
    global FPAM_H_CT, FPAM_V_CT, FSAM_H_CT, FSAM_V_CT
    del FPAM_H_CT, FPAM_V_CT, FSAM_H_CT, FSAM_V_CT
    global n_radii, n_azimuths, max_angle
    del n_radii, n_azimuths, max_angle
    global cfam_name
    del cfam_name
    # CT and coronagraphic datasets
    global dataset_ct, dataset_ct_syn, dataset_ct_interp, dataset_psf_eq_rad
    del dataset_ct, dataset_ct_syn, dataset_ct_interp, dataset_psf_eq_rad
    global dataset_cor
    del dataset_cor
    # Arbitrary set of PSF locations to be tested in EXCAM pixels referred to (0,0)
    global psf_loc_in, psf_loc_syn
    del psf_loc_in, psf_loc_syn
    # CT values
    global ct_in, ct_syn
    del ct_in, ct_syn

if __name__ == '__main__':
    setup_module()
    test_psf_pix_and_ct()
    test_fpm_pos()
    test_cal_file()
    test_ct_interp()
    test_get_1d_ct()
    test_ct_map()
    test_psf_interp()



"""
Calibration tracking system. Modified from kpicdrp caldb implmentation (Copyright (c) 2024, KPIC Team)
"""
import os
import numpy as np
import pandas as pd
import corgidrp
import corgidrp.data as data
import astropy.time as time

column_dtypes = {
    "Filepath": str,
    "Type": str,
    "MJD": float,
    "EXPTIME": float,
    "Files Used": int,
    "Date Created": float,
    "Hash": str,
    "DRPVERSN": str,
    "OBSNUM": int,
    "NAXIS1": int,
    "NAXIS2": int,
    "OPMODE": str,
    "EMGAIN_C": float,
    "EXCAMT": float
}

column_names = list(column_dtypes.keys())

labels = {data.Dark: "Dark",
          data.NonLinearityCalibration: "NonLinearityCalibration",
          data.KGain : "KGain",
          data.BadPixelMap: "BadPixelMap",
          data.DetectorNoiseMaps: "DetectorNoiseMaps",
          data.FlatField : "FlatField",
          data.DetectorParams : "DetectorParams",
          data.AstrometricCalibration : "AstrometricCalibration",
          data.TrapCalibration : "TrapCalibration",
          data.FluxcalFactor : "FluxcalFactor",
          data.FpamFsamCal : "FpamFsamCal",
          data.CoreThroughputCalibration: "CoreThroughputCalibration",
          data.NDFilterSweetSpotDataset: "NDFilterSweetSpot",
          }

class CalDB:
    """
    Database for tracking calibration files saved to disk. Modified from the kpicdrp version

    Note that database is not parallelism-safe, but should be ok in most cases.
    (Jason: look at using posix_ipc to guarantee thread safety if we really need it)

    Args:
        filepath (str): [optional] filepath to a CSV file with an existing database

    Fields:
        columns (list): column names of dataframe
        filepath(str): full filepath to data
    """

    def __init__(self, filepath=""):
        """
        Args:
            filepath (str): [optional] filepath to a CSV file with an existing database
        """

        # If filepath is not passed in, use the default (majority of case)
        if len(filepath) == 0:
            self.filepath = corgidrp.caldb_filepath
        else:
            # possibly edge case where we want to use specialized caldb
            self.filepath = filepath

        # if database does't exist, create a blank one
        if not os.path.exists(self.filepath):
            # new database
            self.columns = column_names
            self._db = pd.DataFrame(columns=self.columns)
            self.save()
        else:
            # already a database exists
            self.load()
            self.columns = list(self._db.columns.values)

    def load(self):
        """
        Load/update db from filepath
        """
        self._db = pd.read_csv(self.filepath, dtype=column_dtypes)

    def save(self):
        """
        Save file without numbered index to disk with user specified filepath as a CSV file
        """
        self._db.to_csv(self.filepath, index=False)

    def _get_values_from_entry(self, entry, is_calib=True):
        """
        Extract the properties from this data entry to ingest them into the database

        Args:
            entry (corgidrp.data.Image subclass): calibration frame to add to the database
            is_calib (bool): is a calibration frame. if Not, it won't look up filetype.
                             Only used in get_calib() to grab metadata for science frames

        Returns:
            tuple:
                row (list):
                    List of data entry properties
                row_dict (dict):
                    Dictionary of data entry properties keyed by column names

        """
        # return a dummy entry if nothing is passed in
        if entry is None:
            time_now = time.Time.now()
            row_dict = {
                "Filepath" : "",
                "Type" : "Sci",
                "MJD" : time_now.mjd,
                "EXPTIME" : 0.,
                "Files Used" : 0,
                "Date Created" : time_now.mjd,
                "Hash" : hash(time_now),
                "DRPVERSN" : "0.0",
                "OBSNUM" : 000,
                "NAXIS1": 0,
                "NAXIS2" : 0,
                "OPMODE" : "",
                "EMGAIN_C" : 0.,
                "EXCAMT" : 0
            }
            return list(row_dict.values()), row_dict

        filepath = os.path.abspath(entry.filepath)
        if is_calib:
            datatype = labels[entry.__class__]  # get the database str representation
        else:
            datatype = "Sci"
        mjd = time.Time(entry.ext_hdr["SCTSRT"]).mjd
        exptime = entry.ext_hdr["EXPTIME"]

        # check if this exists. will be a keyword written by corgidrp
        if "DRPNFILE" in entry.ext_hdr:
            files_used = entry.ext_hdr["DRPNFILE"]
        else:
            files_used = 0

        if "DRPCTIME" in entry.ext_hdr:
            date_created = time.Time(entry.ext_hdr["DRPCTIME"]).mjd
        else:
            date_created = -1

        if "DRPVERSN" in entry.ext_hdr:
            drp_version = entry.ext_hdr["DRPVERSN"]
        else:
            drp_version = ""

        obsid = entry.pri_hdr["OBSNUM"]

        hash_val = entry.get_hash()

        # this only works for 2D images. may need to adapt for non-2D calibration frames
        # import IPython; IPython.embed()

        entry_shape = entry.data.shape
        if len(entry_shape) < 2:
            naxis1 = entry.data.shape[-1]
            naxis2 = 0
        else:
            naxis1 = entry.data.shape[-1]
            naxis2 = entry.data.shape[-2]

        # naxis1 = entry.data.shape[-1]
        # naxis2 = entry.data.shape[-2]

        row = [
            filepath,
            datatype,
            mjd,
            exptime,
            files_used,
            date_created,
            hash_val,
            drp_version,
            obsid,
            naxis1,
            naxis2
        ]

        # rest are ext_hdr keys we can copy
        start_index = len(row)
        for i in range(start_index, len(self.columns)):
            row.append(entry.ext_hdr[self.columns[i]])  # add value staright from header

        row_dict = {}
        for key, val in zip(self.columns, row):
            row_dict[key] = val

        return row, row_dict

    def create_entry(self, entry, to_disk=True):
        """
        Add a new entry to or update an existing one in the database. Note that function by default will load and save db to disk

        Args:
            entry (corgidrp.data.Image subclass): calibration frame to add to the database
            to_disk (bool): True by default, will update DB from disk before adding entry and saving it back to disk
        """
        new_row, row_dict = self._get_values_from_entry(entry)

        # update database from disk in case anything changed
        if to_disk:
            self.load()

        # use filepath as key to see if it's already in database
        if row_dict["Filepath"] in self._db.values:
            row_index = self._db[
                self._db["Filepath"] == row_dict["Filepath"]
            ].index.values
            self._db.loc[row_index, self.columns] = new_row
        # otherwise create new entry
        else:
            new_entry = pd.DataFrame([new_row], columns=self.columns)
            if len(self._db) == 0:
                self._db = new_entry
            else:
                self._db = pd.concat([self._db, new_entry], ignore_index=True)

        # save to disk to update changes
        if to_disk:
            self.save()

    def remove_entry(self, entry, to_disk=True):
        """
        Remove an entry from the database. Removes the entire row

        Args:
            entry (corgidrp.data.Image subclass): calibration frame to add to the database
            to_disk (bool): True by default, will update DB from disk before adding entry and saving it back to disk
        """
        new_row, row_dict = self._get_values_from_entry(entry)

        # update database from disk in case anything changed
        if to_disk:
            self.load()

        if row_dict["Filepath"] in self._db.values:
            entry_index = self._db[
                self._db["Filepath"] == row_dict["Filepath"]
            ].index.values
            self._db = self._db.drop(self._db.index[entry_index])
            self._db = self._db.reset_index(drop=True)
        else:
            raise ValueError("No filepath found so could not remove.")

        # save to disk to update changes
        if to_disk:
            self.save()

    def get_calib(self, frame, dtype, to_disk=True):
        """
        Outputs the best calibration file of the given type for the input science frame.

        Args:
            frame (corgidrp.data.Image): an image frame to request a calibration for. If None is passed in, looks for the 
                                         most recently created calibration. 
            dtype (corgidrp.data Class): for example: corgidrp.data.Dark (TODO: document the entire list of options)
            to_disk (bool): True by default, will update DB from disk before matching

        Returns:
            corgidrp.data.*: an instance of the appropriate calibration type (Exact type depends on calibration type)
        """
        if dtype not in labels:
            raise ValueError(
                "Requested calibration dtype of {0} not a valid option".format(dtype)
            )
        dtype_label = labels[dtype]

        # get values for this science frame
        _, frame_dict = self._get_values_from_entry(frame, is_calib=False)

        # update database from disk in case anything changed
        if to_disk:
            self.load()

        # downselect to only calibs of this type
        calibdf = self._db[self._db["Type"] == dtype_label]

        # different logic for different cases
        # each if/else statement returns a single filepath to a good calibration
        if frame is None:
            # no frame is passed in, get the most recently created 
            options = calibdf

            if len(options) == 0:
                raise ValueError("No valid {0} calibration in caldb located at {1}".format(dtype_label, self.filepath))

            # select the one that was most recently created
            result_index = options["Date Created"].argmax()
            calib_filepath = options.iloc[result_index, 0]

        elif dtype_label in ["Dark"]:
            # general selection criteria for 2D image frames. Can use different selection criteria for different dtypes
            options = calibdf.loc[
                (
                    (calibdf["EXPTIME"] == frame_dict["EXPTIME"])
                )
            ]

            if len(options) == 0:
                raise ValueError("No valid Dark with EXPTIME={0})"
                                 .format(frame_dict["EXPTIME"]))

            # select the one closest in time
            result_index = np.abs(options["MJD"] - frame_dict["MJD"]).argmin()
            calib_filepath = options.iloc[result_index, 0]
        else:
            options = calibdf

            if len(options) == 0:
                raise ValueError("No valid {0} calibration in caldb located at {1}".format(dtype_label, self.filepath))

            # select the one closest in time
            result_index = np.abs(options["MJD"] - frame_dict["MJD"]).argmin()
            calib_filepath = options.iloc[result_index, 0]

        # load the object from disk and return it
        return dtype(calib_filepath)
    
    def scan_dir_for_new_entries(self, filedir, look_in_subfolders=True, to_disk=True):
        """
        Scan a folder and subfolder for calibration files and add them all to the caldb

        Args:
            filedir (str): path to folder to scan (includes all subfolders by default)
            look_in_subfolders (bool): whether to look in subfolders for files. True by default
            to_disk (bool): True by default, will update DB from disk before adding entry and saving it back to disk
        """
        calib_frames = []
        # walk the directory to find all the calibration files
        for dirpath, subfolders, filenames in os.walk(filedir):
            for filename in filenames:
                # hard coded check only for files that end in .fits
                if filename[-5:] != ".fits":
                    continue

                filepath = os.path.join(dirpath, filename)
                frame = data.autoload(filepath)

                # check what class it has been loaded as. only save frames that fall into calibration classes
                if frame.__class__ in labels:
                    calib_frames.append(frame)

            # the first iteration looks in the basedir
            # if we don't wnat to look in subdirs now, we should break
            if not look_in_subfolders:
                break

        # load all these files into the caldb
        for calib_frame in calib_frames:
            self.create_entry(calib_frame, to_disk=to_disk)



def initialize():
    """
    Creates default calibrations and caldb if it doesn't exist

    """
    global initialized

    ### Create set of default calibrations
    rescan_needed = False
    # Add default detector_params calibration file if it doesn't exist
    if not os.path.exists(os.path.join(corgidrp.default_cal_dir, "DetectorParams_2023-11-01T00:00:00.000.fits")):
        default_detparams = data.DetectorParams({}, date_valid=time.Time("2023-11-01 00:00:00", scale='utc'))
        default_detparams.save(filedir=corgidrp.default_cal_dir)
        rescan_needed = True
    # Add default FpamFsamCal calibration file if it doesn't exist
    if not os.path.exists(os.path.join(corgidrp.default_cal_dir, "FpamFsamCal_2024-02-10T00:00:00.000.fits")):
        fpamfsam_2excam = data.FpamFsamCal([],
            date_valid=time.Time("2024-02-10 00:00:00", scale='utc'))
        fpamfsam_2excam.save(filedir=corgidrp.default_cal_dir)
        rescan_needed = True

    if rescan_needed:
        # add default caldb entries
        default_caldb = CalDB()
        default_caldb.scan_dir_for_new_entries(corgidrp.default_cal_dir)

    # set initialization
    initialized = True

initialized = False
initialize()


#A file to test the non-linearity correction, including a comparison with the II&T pipeline
import os
import glob
import pickle
import numpy as np
import corgidrp.mocks as mocks
import corgidrp.data as data
import corgidrp.l1_to_l2a as l1_to_l2a
from scipy import interpolate
import pytest

class NonlinException(Exception):
    """Exception class for nonlin module."""

def _parse_file(nonlin_path):
    """

    ** This function has been copied directly from the II&T pipeline 
    and is only used here to validate the new DRP vs. the old pipeline **

    Get data from nonlin file.
    
    
    Args:
        nonlin_path (str): (Optional) Full path to the non-linearity calibrationf ile.

    Returns:
        tuple:
            gain_ax (numpy.array): The gain axis
            count_ax (numpy.array): The count axis
            relgains (numpy.array): The relative gains
    """
    # Read nonlin csv
    nonlin_raw = np.genfromtxt(nonlin_path, delimiter=',')

    # File format checks
    if nonlin_raw.ndim < 2 or nonlin_raw.shape[0] < 2 or \
       nonlin_raw.shape[1] < 2:
        raise NonlinException('Nonlin array must be at least 2x2 (room for x '
                              'and y axes and one data point)')
    if not np.isnan(nonlin_raw[0, 0]):
        raise NonlinException('First value of csv (upper left) must be set to '
                              '"nan"')

    # Column headers are gains, row headers are dn counts
    gain_ax = nonlin_raw[0, 1:]
    count_ax = nonlin_raw[1:, 0]
    # Array is relative gain values at a given dn count and gain
    relgains = nonlin_raw[1:, 1:]

    # Check for increasing axes
    if np.any(np.diff(gain_ax) <= 0):
        raise NonlinException('Gain axis (column headers) must be increasing')
    if np.any(np.diff(count_ax) <= 0):
        raise NonlinException('Counts axis (row headers) must be increasing')
    # Check that curves (data in columns) contain or straddle 1.0
    if (np.min(relgains, axis=0) > 1).any() or \
       (np.max(relgains, axis=0) < 1).any():
        raise NonlinException('Gain curves (array columns) must contain or '
                              'straddle a relative gain of 1.0')

    return gain_ax, count_ax, relgains


def get_relgains(frame, em_gain, nonlin_path):
    """

    ** This function has been copied directly from the II&T pipeline 
    and is only used here to validate the new DRP vs. the old pipeline 
    It is different from the function in detector.py**

    For a given bias subtracted frame of dn counts, return a same sized
    array of relative gain values.

    Args:
        frame (np.array): Array of dn count values.
        em_gain (float): Detector EM gain.
        nonlin_path (str): Full path of nonlinearity calibration csv.

    Returns:
        array_like : Array of relative gain values.
            

    Notes:
        This algorithm contains two interpolations:

         - A 2d interpolation to find the relative gain curve for a given EM gain
         - A 1d interpolation to find a relative gain value for each given dn
         count value.

        Both of these interpolations are linear, and both use their edge values as
        constant extrapolations for out of bounds values.
    """

    # Get file data
    gain_ax, count_ax, relgains = _parse_file(nonlin_path)

    # Create interpolation for em gain (x), counts (y), and relative gain (z).
    # Note that this defaults to using the edge values as fill_value for
    # out of bounds values (same as specified below in interp1d)
    f = interpolate.RectBivariateSpline(gain_ax,
                                    count_ax,
                                    relgains.T,
                                    kx=1,
                                    ky=1,
    )
    # Get the relative gain curve for the given gain value
    relgain_curve = f(em_gain, count_ax)[0]

    # Create interpolation for dn counts (x) and relative gains (y). For
    # out of bounds values use edge values
    ff = interpolate.interp1d(count_ax, relgain_curve, kind='linear',
                              bounds_error=False,
                              fill_value=(relgain_curve[0], relgain_curve[-1]))
    # For each dn count, find the relative gain
    counts_flat = ff(frame.ravel())

    return counts_flat.reshape(frame.shape)

def test_non_linearity_correction():
    """
    Generate a non-linearity correction calibration and test the correction 
    
    Ported from II&T Pipeline
    """

    #Create a mock dataset because it is a required input when creating a NonLinearityCalibration
    dummy_dataset = mocks.create_prescan_files()

    # Make a non-linearity correction calibration file
    input_non_linearity_filename = "nonlin_table_TVAC.txt"
    input_non_linearity_path = os.path.join(os.path.dirname(__file__), "test_data", input_non_linearity_filename)
    test_non_linearity_filename = input_non_linearity_filename.split(".")[0] + ".fits"
    test_non_linearity_path = os.path.join(os.path.dirname(__file__), "test_data", test_non_linearity_filename)
    tvac_nonlin_data = np.genfromtxt(input_non_linearity_path, delimiter=",")


    pri_hdr, ext_hdr = mocks.create_default_L1_headers()
    non_linearity_correction = data.NonLinearityCalibration(tvac_nonlin_data,pri_hdr=pri_hdr,ext_hdr=ext_hdr,input_dataset = dummy_dataset)
    non_linearity_correction.save(filename = test_non_linearity_path)

    # check the nonlin can be pickled (for CTC operations)
    pickled = pickle.dumps(non_linearity_correction)
    pickled_nonlin = pickle.loads(pickled)
    assert np.all((non_linearity_correction.data == pickled_nonlin.data) | np.isnan(non_linearity_correction.data))

    # import IPython; IPython.embed()

    ###### create a simulated dataset that is non-linear
    # check that simulated data folder exists, and create if not
    datadir = os.path.join(os.path.dirname(__file__), "simdata")
    if not os.path.exists(datadir):
        os.mkdir(datadir)

    emgain = 2000
    mocks.create_nonlinear_dataset(test_non_linearity_path, filedir=datadir,em_gain=emgain)

    ####### open up the files
    sim_data_filenames = glob.glob(os.path.join(datadir, "simcal_nonlin*.fits"))
    nonlinear_dataset = data.Dataset(sim_data_filenames)
    assert len(nonlinear_dataset) == 2

    ######## perform non-linearity correction
    non_linearity_correction = data.NonLinearityCalibration(test_non_linearity_path)

    # check the nonlin can be pickled (for CTC operations)
    pickled = pickle.dumps(non_linearity_correction)
    pickled_nonlin = pickle.loads(pickled)
    assert np.all((non_linearity_correction.data == pickled_nonlin.data) | np.isnan(non_linearity_correction.data))

    # set up values for testing flags and the value of the flag
    non_linear_flag = 64
    non_linear_pixel_value = 1e12

    # seperate dataset to test flagging
    flagged_dataset = nonlinear_dataset.copy()

    # Inject a non-linear pixel and check that if it is flagged
    flagged_dataset.all_data[0,0,0] = non_linear_pixel_value
    flagged_dataset = l1_to_l2a.correct_nonlinearity(flagged_dataset, non_linearity_correction,threshold=non_linear_pixel_value-1)

    assert flagged_dataset.all_dq[0,0,0] >= non_linear_flag # flagged_dataset.all_data[0,0,0] should be flagged
    flagged_dataset.all_dq[0,0,0] = 0 # reset the flag

    assert np.all(flagged_dataset.all_dq < non_linear_flag) # all other pixels should not be flagged 

    linear_dataset = l1_to_l2a.correct_nonlinearity(nonlinear_dataset, non_linearity_correction)

    #The data was generated with a ramp in the x-direction going from 10 to 65536
    expected_ramp = np.linspace(800,65536,1024)
    #Let's collapse the data and see if there's a ramp. 
    collapsed_data = np.mean(linear_dataset.all_data, axis=(0,1))

    #Relative correction
    relative_correction = (collapsed_data-expected_ramp)/collapsed_data

    #We are happy if the relative correction is less than 1% [TBC]
    assert np.all(relative_correction < 1e-2)


    #Let's test that this returns the same thing as the II&T pipeline
    linear_data_iit = nonlinear_dataset.all_data*get_relgains(nonlinear_dataset.all_data,emgain,input_non_linearity_path)

    #We want the difference between the II&T version and ours to be zero. 
    assert np.all(np.abs(linear_dataset.all_data[0]-linear_data_iit[0]) < 1e-5)
    # assert np.mean(linear_dataset.all_data - linear_data_iit) == pytest.approx(0, abs=1e-2)

if __name__ == "__main__":
    test_non_linearity_correction()


# A file that holds the functions that transmogrify l1 data to l2a data
from corgidrp.detector import get_relgains, slice_section, detector_areas, flag_cosmics, calc_sat_fwc, imaging_slice, imaging_area_geom
import numpy as np
import corgidrp.data as data

def prescan_biassub(input_dataset, noise_maps=None, return_full_frame=False, 
                    detector_regions=None, use_imaging_area = False):
    """
    Measure and subtract the median bias in each row of the pre-scan detector region.
    This step also crops the images to just the science area, or
    optionally returns the full detector frames.

    Args:
        input_dataset (corgidrp.data.Dataset): a dataset of Images (L1a-level)
        noise_maps (corgidrp.data.DetectorNoiseMaps): the bias offset (an offset value to be subtracted from the bias) is extracted from this calibration class instance.
        If None, a default value of 0 is used for the bias offset and its error.
        return_full_frame (bool): flag indicating whether to return the full frame or
            only the bias-subtracted image area. Defaults to False.
        detector_regions: (dict):  A dictionary of detector geometry properties.
            Keys should be as found in detector_areas in detector.py. Defaults to detector_areas in detector.py.
        use_imaging_area (bool): flag indicating whether to use the imaging area (like in the trap pump code) or use the defualt (equivalent to EMCCDFrame)

    Returns:
        corgidrp.data.Dataset: a pre-scan bias subtracted version of the input dataset
    """

    if detector_regions is None:
        detector_regions = detector_areas

    # Initialize list of output frames to be concatenated
    out_frames_data = []
    out_frames_err = []
    out_frames_dq = []
    out_frames_bias = []

    # Place to save new error estimates to be added later via Image.add_error_term()
    new_err_list = []

    # Iterate over frames
    for i, frame in enumerate(input_dataset):

        frame_data = np.copy(frame.data)
        frame_err = np.copy(frame.err)
        frame_dq = np.copy(frame.dq)

        # Determine what type of file it is (engineering or science), then choose detector area dict
        arrtype = frame.ext_hdr['ARRTYPE']
        if not arrtype in ['SCI','ENG','ENG_EM','ENG_CONV'] :
                raise Exception(f"Observation type of frame {i} is not 'SCI' or 'ENG' or 'ENG_EM' or 'EMG_CONV'")

        if detector_regions[arrtype]['frame_rows'] != frame_data.shape[0] or detector_regions[arrtype]['frame_cols'] != frame_data.shape[1]:
            raise Exception('Frame size incompatible with specified detector_regions.')
        # Get the reliable prescan area
        prescan = slice_section(frame_data, arrtype, 'prescan', detector_regions=detector_regions)

        if not return_full_frame:
            # Get the image area
            if use_imaging_area: 
                image_data = imaging_slice(arrtype, frame_data, detector_regions=detector_regions)
                image_dq = imaging_slice(arrtype, frame_dq, detector_regions=detector_regions)

                image_err = []
                for err_slice in frame_err:
                    image_err.append(imaging_slice(arrtype, err_slice, detector_regions=detector_regions))
                image_err = np.array(image_err)

                prows, _, r0c0 = imaging_area_geom(arrtype,detector_regions=detector_regions)
                i_r0 = r0c0[0]
                p_r0 = detector_regions[arrtype]['prescan']['r0c0'][0]
                al_prescan = prescan[(i_r0-p_r0):(i_r0-p_r0+prows), :]

            else: 
                image_data = slice_section(frame_data, arrtype, 'image', detector_regions)
                image_dq = slice_section(frame_dq, arrtype, 'image', detector_regions)

                # Special treatment for 3D error array
                image_err = []
                for err_slice in frame_err:
                    image_err.append(slice_section(err_slice, arrtype, 'image', detector_regions))
                image_err = np.array(image_err)

                # Get the part of the prescan that lines up with the image
                i_r0 = detector_areas[arrtype]['image']['r0c0'][0]
                p_r0 = detector_areas[arrtype]['prescan']['r0c0'][0]
                i_nrow = detector_areas[arrtype]['image']['rows']
                al_prescan = prescan[(i_r0-p_r0):(i_r0-p_r0+i_nrow), :]

        else:
            # Use full frame
            image_data = frame_data
            image_dq = frame_dq

            # Special treatment for 3D error array
            image_err = []
            for err_slice in frame_err:
                image_err.append(err_slice)
            image_err = np.array(image_err)

            al_prescan = prescan

        st = detector_regions[arrtype]['prescan']['col_start']
        end = detector_regions[arrtype]['prescan']['col_end']

        # Measure bias and error (standard error of the median for each row, add this to 3D image array)
        medbyrow = np.median(al_prescan[:,st:end], axis=1)[:, np.newaxis]
        sterrbyrow = np.std(al_prescan[:,st:end], axis=1)[:, np.newaxis] * np.ones_like(image_data) / np.sqrt(al_prescan[:,st:end].shape[1])
        if noise_maps is not None:
            bias_offset = noise_maps.bias_offset
            bias_offset_err = noise_maps.bias_offset_err
        else:
            bias_offset = 0
            bias_offset_err = 0
        sterrbyrow = np.sqrt(sterrbyrow**2 + bias_offset_err**2)
        new_err_list.append(sterrbyrow)

        bias = medbyrow - bias_offset
        image_bias_corrected = image_data - bias

        out_frames_data.append(image_bias_corrected)
        out_frames_err.append(image_err)
        out_frames_dq.append(image_dq)
        out_frames_bias.append(bias[:,0]) # save 1D version of array

    # Update all_data and reassign frame pointers (only necessary because the array size has changed)
    out_frames_data_arr = np.array(out_frames_data)
    out_frames_err_arr = np.array(out_frames_err)
    out_frames_dq_arr = np.array(out_frames_dq)
    out_frames_bias_arr = np.array(out_frames_bias, dtype=np.float32)
    # try to free some memory
    del out_frames_data, out_frames_err, out_frames_dq, out_frames_bias

    # Make a copy of the input dataset to operate on
    output_dataset = input_dataset.copy(copy_data=False)

    output_dataset.all_data = out_frames_data_arr
    output_dataset.all_err = out_frames_err_arr
    output_dataset.all_dq = out_frames_dq_arr

    for i,frame in enumerate(output_dataset):
        frame.data = out_frames_data_arr[i]
        frame.err = out_frames_err_arr[i]
        frame.dq = out_frames_dq_arr[i]
        # frame.bias = out_frames_bias_arr[i]
        frame.add_extension_hdu("BIAS",data=out_frames_bias_arr[i])

        # Update header with new frame dimensions
        frame.ext_hdr['NAXIS1'] = out_frames_data_arr[i].shape[1]
        frame.ext_hdr['NAXIS2'] = out_frames_data_arr[i].shape[0]

    # Add new error component from this step to each frame using the Dataset class method
    output_dataset.add_error_term(np.array(new_err_list),"prescan_bias_sub")

    history_msg = "Frames cropped and bias subtracted" if not return_full_frame else "Bias subtracted"

    # update the output dataset with this new dark subtracted data and update the history
    output_dataset.update_after_processing_step(history_msg)

    return output_dataset

def detect_cosmic_rays(input_dataset, detector_params, k_gain = None, sat_thresh=0.7,
                       plat_thresh=0.7, cosm_filter=1, cosm_box=3, cosm_tail=10,
                       mode='image', detector_regions=None):
    """
    Detects cosmic rays in a given dataset. Updates the DQ to reflect the pixels that are affected.
    TODO: (Eventually) Decide if we want to invest time in improving CR rejection (modeling and subtracting the hit
    and tail rather than just flagging the whole row.)
    TODO: Decode incoming DQ mask to avoid double counting saturation/CR flags in case a similar custom step has been run beforehand.

    Args:
        input_dataset (corgidrp.data.Dataset): a dataset of Images that need cosmic ray identification (L1-level)
        detector_params (corgidrp.data.DetectorParams): a calibration file storing detector calibration values
        k_gain (corgidrp.data.KGain): KGain calibration file
        sat_thresh (float):
            Multiplication factor for the pixel full-well capacity (fwc) that determines saturated cosmic
            pixels. Interval 0 to 1, defaults to 0.7. Lower numbers are more aggressive in flagging saturation.
        plat_thresh (float):
            Multiplication factor for pixel full-well capacity (fwc) that determines edges of cosmic
            plateau. Interval 0 to 1, defaults to 0.7. Lower numbers are more aggressive in flagging cosmic
            ray hits.
        cosm_filter (int):
            Minimum length in pixels of cosmic plateaus to be identified. Defaults to 1.
        cosm_box (int):
            Number of pixels out from an identified cosmic head (i.e., beginning of
            the plateau) to mask out.
            For example, if cosm_box is 3, a 7x7 box is masked,
            with the cosmic head as the center pixel of the box. Defaults to 3.
        cosm_tail (int):
            Number of pixels in the row downstream of the end of a cosmic plateau
            to mask.  If cosm_tail is greater than the number of
            columns left to the end of the row from the cosmic
            plateau, the cosmic masking ends at the end of the row. Defaults to 10.
        mode (string):
            If 'image', an image-area input is assumed, and if the input
            tail length is longer than the length to the end of the image-area row,
            the mask is truncated at the end of the row.
            If 'full', a full-frame input is assumed, and if the input tail length
            is longer than the length to the end of the full-frame row, the masking
            continues onto the next row.  Defaults to 'image'.
        detector_regions: (dict):  
            A dictionary of detector geometry properties.  Keys should be as 
            found in detector_areas in detector.py. Defaults to detector_areas in detector.py.

    Returns:
        corgidrp.data.Dataset:
            A version of the input dataset of the input dataset where the cosmic rays have been identified.
    """
    sat_dqval = 32 # DQ value corresponding to full well saturation
    cr_dqval = 128 # DQ value corresponding to CR hit

    if detector_regions is None:
        detector_regions = detector_areas

    # you should make a copy the dataset to start
    crmasked_dataset = input_dataset.copy()

    crmasked_cube = crmasked_dataset.all_data

    # Calculate the full well capacity for every frame in the dataset
    if k_gain is None:
        kgain = detector_params.params['KGAINPAR']
    else:
        #get the kgain value from the k_gain calibration file
        kgain = k_gain.value
    emgain_list = []
    for frame in crmasked_dataset:
        try: # use measured gain if available TODO change hdr name if necessary
            emgain = frame.ext_hdr['EMGAIN_M']
        except:
            if frame.ext_hdr['EMGAIN_A'] > 0: # use applied EM gain if available
                emgain = frame.ext_hdr['EMGAIN_A']
            else: # otherwise use commanded EM gain
                emgain = frame.ext_hdr['EMGAIN_C']
            emgain_list.append(emgain)
    emgain_arr = np.array(emgain_list)
    fwcpp_e_arr = np.array([detector_params.params['FWC_PP_E'] for frame in crmasked_dataset])
    fwcem_e_arr = np.array([detector_params.params['FWC_EM_E'] for frame in crmasked_dataset])

    fwcpp_dn_arr = fwcpp_e_arr / kgain
    fwcem_dn_arr = fwcem_e_arr / kgain

    # pick the FWC that will get saturated first, depending on gain
    sat_fwcs = calc_sat_fwc(emgain_arr,fwcpp_dn_arr,fwcem_dn_arr,sat_thresh)

    for i,frame in enumerate(crmasked_dataset):
        frame.ext_hdr['FWC_PP_E'] = fwcpp_e_arr[i]
        frame.ext_hdr['FWC_EM_E'] = fwcem_e_arr[i]
        frame.ext_hdr['SAT_DN'] = sat_fwcs[i]

    sat_fwcs_array = np.array([np.full_like(crmasked_cube[0],sat_fwcs[i]) for i in range(len(sat_fwcs))])

    # threshold the frame to catch any values above sat_fwc --> this is
    # mask 1
    m1 = (crmasked_cube >= sat_fwcs_array) * sat_dqval

    # run remove_cosmics() with fwc=fwc_em since tails only come from
    # saturation in the gain register --> this is mask 2
    # Do a for loop since it's calling a for loop in the sub-routine anyway
    # and can't handle different 'FWC_EM's for different frames.
    m2 = np.zeros_like(crmasked_cube)

    for i in range(len(crmasked_cube)):
        arrtype = crmasked_dataset.frames[i].ext_hdr['ARRTYPE']
        m2[i,:,:] = flag_cosmics(cube=crmasked_cube[i:i+1,:,:],
                        fwc=fwcem_dn_arr[i],
                        sat_thresh=sat_thresh,
                        plat_thresh=plat_thresh,
                        cosm_filter=cosm_filter,
                        cosm_box=cosm_box,
                        cosm_tail=cosm_tail,
                        mode=mode,
                        detector_regions=detector_regions,
                        arrtype=arrtype
                        ) * cr_dqval

    # add the two masks to the all_dq mask
    new_all_dq = np.bitwise_or(crmasked_dataset.all_dq, m1)
    new_all_dq =  np.bitwise_or(new_all_dq, m2.astype(int))

    history_msg = ("Cosmic ray mask created. "
                   "Used detector parameters from {0}"
                   "with hash {1}").format(detector_params.filename, detector_params.get_hash())

    # update the output dataset with this new dark subtracted data and update the history
    crmasked_dataset.update_after_processing_step(history_msg, new_all_dq=new_all_dq)

    return crmasked_dataset

def correct_nonlinearity(input_dataset, non_lin_correction, threshold=np.inf):
    """
    Perform non-linearity correction of a dataset using the corresponding non-linearity correction. We check for non-linear pixel and flag them in the DQ. 

    Args:
        input_dataset (corgidrp.data.Dataset): a dataset of Images that need non-linearity correction (L2a-level).
        non_lin_correction (corgidrp.data.NonLinearityCorrection): a NonLinearityCorrection calibration file to model the non-linearity.
        threshold (float): threshold for flagging pixels in the DQ array, value above this threshold will be flagged in the DQ map as too nonlinear. By default it is set to infinity, user can change it to a different value.
    
    Returns:
        (corgidrp.data.Dataset): A non-linearity corrected version of the input dataset
    """
    #Copy the dataset to start
    linearized_dataset = input_dataset.copy()

    #Apply the non-linearity correction to the data
    linearized_cube = linearized_dataset.all_data

    #Check to see if EM gain is in the header, if not, raise an error
    if "EMGAIN_C" not in linearized_dataset[0].ext_hdr.keys():
        raise ValueError("EM gain not found in header of input dataset. Non-linearity correction requires EM gain to be in header.")

    for i in range(linearized_cube.shape[0]):
        try: # use measured gain if available TODO change hdr name if necessary
            em_gain = linearized_dataset[i].ext_hdr["EMGAIN_M"]
        except:
            em_gain = linearized_dataset[i].ext_hdr["EMGAIN_A"]
            if em_gain > 0: # use applied EM gain if available
                em_gain = linearized_dataset[i].ext_hdr["EMGAIN_A"]
            else: # otherwise use commanded EM gain
                em_gain = linearized_dataset[i].ext_hdr["EMGAIN_C"]

        # Flag pixels in the DQ array if they exceed the threshold
        non_linear_flag = 64
        current_value = linearized_dataset[i].dq[linearized_cube[i] > threshold]
        linearized_dataset[i].dq[linearized_cube[i] > threshold] = np.bitwise_or(current_value, non_linear_flag)
        linearized_cube[i] *= get_relgains(linearized_cube[i], em_gain, non_lin_correction)
    
    if non_lin_correction is not None:
        history_msg = "Data corrected for non-linearity with {0}".format(non_lin_correction.filename)

        linearized_dataset.update_after_processing_step(history_msg, new_all_data=linearized_cube)

    return linearized_dataset

def update_to_l2a(input_dataset):
    """
    Updates the data level to L2a. Only works on L1 data.

    Currently only checks that data is at the L1 level

    Args:
        input_dataset (corgidrp.data.Dataset): a dataset of Images (L1-level)

    Returns:
        corgidrp.data.Dataset: same dataset now at L2-level
    """
    # check that we are running this on L1 data
    for orig_frame in input_dataset:
        if orig_frame.ext_hdr['DATALVL'] != "L1":
            err_msg = "{0} needs to be L1 data, but it is {1} data instead".format(orig_frame.filename, orig_frame.ext_hdr['DATALVL'])
            raise ValueError(err_msg)

    # we aren't altering the data
    updated_dataset = input_dataset.copy(copy_data=False)

    for frame in updated_dataset:
        # update header
        frame.ext_hdr['DATALVL'] = "L2a"
        # update filename convention. The file convention should be
        # "CGI_[dataleel_*]" so we should be same just replacing the just instance of L1
        frame.filename = frame.filename.replace("_L1_", "_L2a", 1)

    history_msg = "Updated Data Level to L2a"
    updated_dataset.update_after_processing_step(history_msg)

    return updated_dataset



import os
import numpy as np
import corgidrp
from corgidrp.mocks import create_default_L2b_headers
from corgidrp.data import Image, Dataset
from corgidrp.l2a_to_l2b import add_photon_noise
import pytest

old_err_tracking = corgidrp.track_individual_errors

data = np.ones([1024,1024])*2.
err = np.ones([1024,1024]) *0.5
dq = np.zeros([1024,1024], dtype = np.uint16)
# TO DO: Check to confirm this is correct data level
prhd, exthd = create_default_L2b_headers()

def test_add_phot_noise():
    corgidrp.track_individual_errors = True

    print("test add_photon_noise pipeline step")
    image1 = Image(data,pri_hdr = prhd, ext_hdr = exthd, err = err, dq = dq)
    image2 = Image(data,pri_hdr = prhd, ext_hdr = exthd, err = err, dq = dq)
    
    dataset = Dataset([image1, image2])
    dataset_add = add_photon_noise(dataset)
    all_err = dataset.all_err
    all_err1 = dataset_add.all_err
    assert not np.array_equal(all_err, all_err1)
    assert np.array_equal(all_err1[0,1], np.sqrt(data))
    assert np.allclose(all_err1[0,0], np.sqrt(data + np.square(err)))
    assert "noise" in str(dataset_add.frames[0].ext_hdr["HISTORY"])
    assert "photnoise_error" == dataset_add.frames[0].err_hdr["Layer_2"]
    #check that excess noise is applied
    dataset[0].ext_hdr["EMGAIN_C"] = 3000
    dataset_add1 = add_photon_noise(dataset)
    all_err2 = dataset_add1.all_err
    assert np.array_equal(all_err2[0,1], np.sqrt(data)*np.sqrt(2))
    
    corgidrp.track_individual_errors = old_err_tracking
    
if __name__ == '__main__':
    test_add_phot_noise()
